package BSTW;

#interface
use Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(&bstw_encode &bstw_decode &print_help_bstw &bstw_count_bits);

#implementation
use strict;

use Common;

###############################################################################
# BSTW                                                                       ##
###############################################################################

sub print_help_bstw
{
	print "\nBSTW\n\
Metoda BSTW kduje pmo slova pirozenho jazyka. Pokud se m zakdovat slovo, \
kter jet nen ve slovnku, ulo se na zatek a na vstup se pole poadov \
slo (odpovd velikosti slovnku po pidn slova) a dan slovo. V opanm \
ppad se pole na vstup pouze slo, kter odpovd aktulnmu poad slova \
ve slovnku. Nsledn se slovo pesune na zatek slovnku.\
Metoda je vhodn pedevm tam, kde se lokln opakuj stejn slova, protoe \
na vstup se poslaj men sla = krat kdov slova pi pouit vhodnho \
kdovn pirozench sel kdy promnn dlky (nap. gama').\
Zde bylo pouito binrn kdovn pevn dllky (5 bit).
\n\n"; 
}

sub bstw_count_bits
{
# in:
# encoded (ref array)
# out:
# count of bits

	my @encoded = @{shift @_};
	my $bits = 0;
	foreach $a (@encoded) {
		my @r = split /_/, $a;
		# rozlisovaci bit - neni potreba
		if ($#r == 0) { # jenom cislo
			$bits += 5;
		}
		else { # cislo i slovo
 			$bits += 5;
			$bits += length($r[1]) * 8;
		}
	}
	return $bits;
}


# encoding BSTW

sub bstw_encode
{
# in:
# input (ref array)
# mode (scalar)
# out:
# encoded (array)

	my $ref_input = shift @_;
	my $mode = shift;

	my @input = split /\s/, $$ref_input;
	my $input_length = $#input + 1;
	my @encoded = ();
	my @dict = ();

	foreach my $a (@input) {
		my $found = 0;
		my $out;
		for (my $i = 0; $i <= $#dict; $i++) {
			if ($dict[$i] eq $a) {
				$found = 1;
				$out = $i+1;
				my $pom = $dict[$i];
				splice @dict, $i, 1;
				unshift @dict, $pom;
				last;
			}
		}
		if (!$found) {
			$out = ($#dict + 2)."_$a";
			unshift @dict, $a;
		}
		push @encoded, $out;
		
		printf "%-12s =>  %-15s .... dict: %s\n",  $a, translate($out), "@dict"  if no_silent($mode); 
		my_wait($mode);
	}
	return @encoded;
}

# decoding BSTW

sub bstw_decode
{
# in:
# encoded (ref array)
# mode (scalar)
# out:
# decoded (array)

	my $ref_encoded = shift @_;
	my $mode = shift;

	my @encoded = @$ref_encoded;
	my @decoded = ();
	my @dict = ();
	my $dict_size = 0;

	foreach my $a (@encoded) {
		my @r = split /_/, $a;
		my $out;
		if ($r[0] > $dict_size) { #nove slovo
			unshift @dict, $r[1];
			$out = $r[1];
			$dict_size++;
		}
		else {
			$out = @dict[$r[0]-1];
			splice @dict, $r[0]-1, 1;
			unshift @dict, $out;
		}
		push @decoded, $out;
		printf "%-15s =>  %-12s .... dict: %s\n",  translate($a), $out, "@dict"  if no_silent($mode); 
		my_wait($mode);
	}
	return (join ' ', @decoded);
}

1;
