package LZSS;

#interface
use Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(&lzss_encode &lzss_decode &print_help_lzss &lzss_count_bits);

#implementation
use strict;
use POSIX;
use Common;

###############################################################################
# LZSS                                                                       ##
###############################################################################


sub print_help_lzss
{
	print "\nLZSS\n\
Metoda LZSS je velmi podobn metod LZ77. Vyuv tak posuvnho oknka (sliding \
window) o velikosti P+B. Prvn st o velikosti P obsahuje ji zpracovan text \
(nazveme ji processed). Po inicializaci je przdn. Druh st (buffer) obsahuje \
text, kter se bude prv zpracovvat. Na zatku obsahuje zatek vstupu. Jeden \
krok algoritmu vypad tak, e se v poli processed hled posledn vskyt nejdelho \
podetzce zanajcho na zatku bufferu. Je povoleno, aby konec tohoto podetzce \
zasahoval do bufferu. Po nalezen se na vstup posl dvojice (posun, poet), \
kde 'posun' znamen o kolik znak se posouvme zpt a 'poet' - kolik znak se \
m od tto pozice koprovat. Pokud se nenalezne dn shoda, na vstup poslme \
zakdovan znak. Tuto skutenost je teba odliit, a proto poslme na vstup \
rozliovac bit, zda se jedn o kdovou dvojici i nov znak. \
Na konci kroku posuneme oknko tak, aby na zatku bufferu byl prvn nezpracovan znak.\
Pokud byla nalezena jen jednopsmenn shoda, stoj za vahu, zda je lep poslat \
na vstup dvojici i zakdovat psmeno jako nov znak. \
V tomto programu je vhodnj kdovat i jednoznakov shody, protoe je teba \
jen 3+2+1=6 bit, pokud bychom poslali na vstup nov znak, stlo by to 8+1=9 \
bit, co se vyplat a u vt velikosti oknka.\
Pro demonstraci doporuuji pout parametry P = 7, B = 3, piem typick \
hodnoty parametr jsou P <= 8192, B 10-20.\n\n"; 
}

sub lzss_count_bits
{
# in:
# encoded (ref array)
# window_size (ref array)
# out:
# count of bits
	my @encoded = @{shift @_};
	my @window_size = @{shift @_};
	my $bits = 0;
	my $pom = ceil(log($window_size[0])/log(2)) + ceil(log($window_size[1])/log(2));
	foreach $a (@encoded) {
		my @r = split /_/, $a;
		$bits += 1; # rozlisovaci bit
		if ($r[0] == 0) { # novy znak
			$bits += 8;
		}
		else { # posun + pocet
			$bits += $pom;
		}
	}
	return $bits;
}

# encoding LZSS

# kodovad se budou retezce o delce minimalne 2
# moznost nastaveni kodovani retezcu delky jedna

sub lzss_encode
{
# in:
# input (ref array)
# mode (scalar)
# window_size (ref array)
# out:
# encoded (array)

	my $ref_input = shift @_;
	my $mode = shift;
	my @input = split //, $$ref_input;
	my @window_size = @{shift @_};

	my @encoded = ();
	my @processed = (0, -1);
	my @buffer = (0, min($window_size[1], $#input));
	my $input_length = $#input + 1;

	my @ret = find_longest_substring(join('', @input[$buffer[0] .. $buffer[1]]), "");
	my $last = 0;
	while (1) {
		my $character;
		my $ind = $buffer[0] + $ret[1];
		if ($ind >= $input_length) {
			$last = 1;
		}
		
		my $out;
		# zadna shoda - nove pismeno
		if ($ret[1] == 0) {
			# na vystup jen pismeno
			$character = @input[$buffer[0]];
			$out = "0_'$character'";
			push @encoded, $out;
			$ret[1] = 1;
		}
		else {
			$out = "1_$ret[0]_$ret[1]";
			push @encoded, $out;
		}	

		printf "|%*s | %-*s| %s\t=>\t%s\n", 
			2*($window_size[0] + 1), "@input[$processed[0]..$processed[1]]", 
			2*($window_size[1] + 1),  "@input[$buffer[0]..$buffer[1]]",
			defined @input[$buffer[1]+1] ? @input[$buffer[1]+1] : " ",
			translate($out) if no_silent($mode);

		last if $last;

		$buffer[0] += $ret[1];
		$buffer[1] += $ret[1];
		if ($buffer[1] > $input_length - 1) {
			$buffer[1] = $input_length - 1;
		}
		$processed[1] += $ret[1];
		if ($processed[1] > $input_length - 1) {
			$processed[1] = $input_length - 1;
		}
		if ($processed[1] - $processed[0] >= $window_size[0]) {
			$processed[0] = $processed[1] - $window_size[0];
		}
		if ($buffer[0] > $buffer[1]) {
			$last = 1;
		}	
		last if $last;
	
		@ret = find_longest_substring(join('', @input[$buffer[0] .. $buffer[1]]), join('', @input[$processed[0] .. $processed[1]]));
		my_wait($mode);
	}
	return @encoded;
}

# decoding LZSS
sub lzss_decode
{
# in:
# input (ref array)
# mode (scalar)
# window_size (ref array)
# out:
# encoded (array)

	my $ref_encoded = shift @_;
	my $mode = shift;

	my @window_size = @{shift @_};

	my @encoded = @$ref_encoded;
	my @decoded = ();

	foreach my $a (@encoded) {
		my @r = split(/_/, $a);
		if ($r[0] == 1) { # pozice
			my $j = 0;
			for (my $i = $#decoded - $r[1] + 1 ; $j < $r[2]; $i++, $j++) {
				push @decoded, $decoded[$i];
			}
		}
		else { # pismeno
			$r[1] =~ s/'//g;
			push @decoded, $r[1];
		}
		print translate($a) . "\t=>\t@decoded\n" if no_silent($mode);
		my_wait($mode);
	}
	return (join '', @decoded);
}


1;