package LZ77;

#interface
use Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(&lz77_encode &lz77_decode &print_help_lz77 &lz77_count_bits);

#implementation
use strict;
use POSIX;
use Common;

###############################################################################
# LZ77                                                                       ##
###############################################################################

sub print_help_lz77
{
	print "\nLZ77\n\
Metoda LZ77 (angl. sliding window) vyuv oknka o velikosti P+B. Prvn st\
o velikosti P obsahuje ji zpracovan text (nazveme ji processed). Po inicializaci\
je przdn. Druh st (buffer) obsahuje text, kter se bude prv zpracovvat.\
Na zatku obsahuje zatek vstupu. Jeden krok algoritmu vypad tak, e se v poli\
processed hled posledn vskyt nejdelho podetzce zanajcho na zatku \
bufferu. Je povoleno, aby konec tohoto podetzce zasahoval do bufferu.\ 
Po nalezen se na vstup posl trojice (posun, poet, znak), kde 'posun' znamen\
o kolik znak se posouvme zpt, 'poet' - kolik znak se m od tto pozice \
koprovat a 'znak' - jak je dal psmeno, kter ji nebylo obsaeno v podetzci. \
Pot se oknko posune tak, aby 'znak' byl na posledn pozici v poli processed.\n\
Typick hodnoty parametr jsou P <= 8192, B 10-20, pro demonstraci doporuuji \
pout parametry P = 7, B = 3.\n\n"; 

}

sub lz77_count_bits
{
# in:
# encoded (ref array)
# window_size (ref array)
# out:
# count of bits
	my @encoded = @{shift @_};
	my @window_size = @{shift @_};
	my $bits = 0;
	my $tripples = $#encoded + 1;
	$bits = ceil(log($window_size[0])/log(2));
	$bits += ceil(log($window_size[1])/log(2));
	$bits += 8;
	$bits *= $tripples;
	return $bits;
}

sub lz77_encode
{
# in:
# input (ref array)
# mode (scalar)
# window_size (ref array)
# out:
# encoded (array)

	my $ref_input = shift @_;
	my @input = split //, $$ref_input;
	my $mode = shift;
	my @window_size = @{shift @_};

	my @encoded = ();

	my @processed = (0, -1);
	my @buffer = (0, min($window_size[1]-1, $#input));

	my $input_length = $#input + 1;

	my @ret = find_longest_substring(join('', @input[$buffer[0] .. $buffer[1]]), "");
	my $last = 0;
	while (1) {
		my $character;
		my $ind = $buffer[0] + $ret[1];
		if ($ind >= $input_length) {
			$character = '';
			$last = 1;
		}
		else {
			$character = $input[$ind];
		}
		my $out = "$ret[0]_$ret[1]_'$character'";
		push @encoded, $out;

		
		printf "|%*s | %-*s| %s\t=>\t%s\n", 
			2*$window_size[0], "@input[$processed[0]..$processed[1]]", 
			2*$window_size[1],  "@input[$buffer[0]..$buffer[1]]",
			defined @input[$buffer[1]+1] ? @input[$buffer[1]+1] : " ", 
			translate($out) if no_silent($mode);
	
		if ($last) {
			last;
		}
		$buffer[0] += $ret[1] + 1;
		$buffer[1] += $ret[1] + 1;
		if ($buffer[1] > $input_length - 1) {
			$buffer[1] = $input_length - 1;
		}
		$processed[1] += $ret[1] + 1;
		if ($processed[1] > $input_length - 1) {
			$processed[1] = $input_length - 1;
		}
		if ($processed[1] - $processed[0] >= $window_size[0]-1) {
			$processed[0] = $processed[1] - $window_size[0] + 1;
		}
		if ($buffer[0] > $buffer[1]) {
			$last = 1;
		}	
		last if $last;
	
		@ret = find_longest_substring(join('',@input[$buffer[0] .. $buffer[1]]), join('',@input[$processed[0] .. $processed[1]]));
		my_wait($mode);
	}
	return @encoded;
}

# decoding LZ77

sub lz77_decode
{
# in:
# encoded input (ref array)
# mode (scalar)
# window_size (ref array)
# out:
# decoded (array)

	my $ref_encoded = shift @_;
	my $mode = shift;
	my @window_size = @{shift @_};

	my @encoded = @$ref_encoded;

	my @decoded = ();
	foreach my $a (@encoded) {
		my @r = split(/_/, $a);
		if ($r[0] != 0) {
			my $j = 0;
			for (my $i = $#decoded - $r[0] + 1 ; $j < $r[1]; $i++, $j++) {
				push @decoded, $decoded[$i];
			}
		}
		my $end_char = defined $r[2] ? $r[2] : '';
		$end_char =~ s/'//g;
		push @decoded, $end_char;
		print translate($a) . "\t=>\t@decoded\n" if no_silent($mode);
		my_wait($mode);
	}
	return (join '', @decoded);
}

1;