package LZW;

#interface
use Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(&lzw_encode &lzw_decode &print_help_lzw &lzw_count_bits);

#implementation

use strict;
use POSIX;
use Common;

###############################################################################
# LZW                                                                        ##
###############################################################################

sub print_help_lzw
{
	print "\nLZW\n\
Metoda LZW je zaloena na rostoucm slovnku. Text rozkldme zleva na frze tak, \
e prvn psmeno nov frze je poslednm psmenem pedchoz frze. Pidvn \
do slovnku probh stejn jako u metody LZ78. Tedy hledme nejdel prefix nov \
frze, kter je obsaen ve slovnku, prodloume ho o jedno psmeno a pidme frzi \
do slovnku. Slovnk je na zatku inicializovn psmeny z abecedy. Na vstup \
poslme pouze indexy do slovnku.\
Pi naplnn slovnku je mon dle pracovat se stvajcm slovnkem, m se \
me zhorit kompresn pomr. Jin varianty vyuvaj tzv. blokovou kompresi - \
po naplnn se sleduje, jestli se nezhoruje kompresn pomr. Pokud ano, dojde \
k reinicializace slovnku.\
V tomto programu je slovnk implementovn jako strom. Pokud je nastavena volba \
tree_view resp. list_view, po kadm kroku bude vypsn jako strom resp. jako \
seznam (v ppad, e nen nastaven silent_mode). V ppad volby waiting_mode \
lze vypisovat slovnk explicitn psmeny t (tree_view), l (list_view) nebo \
b (both - vype se strom i seznam).\n\n";
}

sub lzw_count_bits
{
# in:
# encoded (ref array)
# ref_alphabet
# out:
# count of bits
	my @encoded = @{shift @_};
	my @alphabet = @{shift @_};
	my $size = $#alphabet;
	my $bits = 0;
	foreach my $code_word (@encoded) {
		my @code_word = split /_/, $code_word;
		my $pom += $code_word[0] == 0 ? 1 : ceil(log($code_word[0])/log(2));
		if ($pom > $size) {
			$size = $pom;
		}	
		$bits += $size;
	}
	return $bits;
}

# encoding LZW

sub get_alphabet
{
# in:
# array of chars
# ref_alphabet
# out:
# array of unique chars

	my $ref_input = shift;
	my $ref_alphabet = shift;
	my %alphabet = ();
	foreach my $a (@$ref_input) {
		$alphabet{$a} = 1;
	}
	@$ref_alphabet = sort keys %alphabet;
}

my $start = 0;

sub lzw_encode
{

# in:
# input (ref array)
# mode (scalar)
# ref_alphabet (empty)
# out:
# encoded (array)

	my $ref_input = shift @_;
	my $mode = shift;
	my $ref_alphabet = shift; # array ref
	my @input = split //, $$ref_input;
	my @encoded = ();

	my $input_length = $#input + 1;

# struct of dictionary:
# node consists of two items: number of node and hash (keys: letters, values: nodes)

	my @dict = ();
	my $counter = $start;
	$dict[0] = $counter;
	
	get_alphabet(\@input, $ref_alphabet);

	my @alphabet = @$ref_alphabet;

	foreach my $a (@alphabet) {
		$dict[1]{$a}[0] = ++$counter; 
		$dict[1]{$a}[1] = {}; 
	}

	print_dict_list(\@dict) if no_silent($mode);
	for (my $i = 0; $i < $input_length; ) {
		my @ret = search_enc_lzw(\@input, \$counter, $i, \@dict, 0);
		my $out = $ret[1];
		push @encoded, $out;
		if ($ret[2] ne '') {
			printf "%2s .. %-6s\t=>\tout: %-6s\t%s\n", "$counter", "@input[$i .. $i+$ret[0]-1]",
			"@input[$i..$i+$ret[0]-2]", "($out)" if no_silent($mode);
			$i += $ret[0] - 1;
		}
		else {
			printf "%2s .. %-6s\t=>\tout: %-6s\t%s\n", "", "", 
			"@input[$i..$i+$ret[0]-1]", "($out)" if no_silent($mode);
			last;
		}
		my $key = my_wait($mode);
		if (defined $key) {
			print_dict_tree(\@dict) if ($key eq 't');
			print_dict_list(\@dict) if ($key eq 'l');
			if ($key eq 'b') {
				print_dict_tree(\@dict);
				print_dict_list(\@dict);
			}
		}
		else {
			print_dict_tree(\@dict) if tree_view($mode);
			print_dict_list(\@dict) if list_view($mode);
		}
	}
	return @encoded;
}

sub search_enc_lzw
{
	my ($ref_input, $ref_counter, $ind, $ref_dict, $depth) = @_;
	my @input = @$ref_input;
	my $input_length = $#input + 1;

	return ($depth, $$ref_dict[0], '') if $ind >= $input_length;
	if (exists $$ref_dict[1]{$input[$ind]}) {
		return search_enc_lzw($ref_input, $ref_counter, $ind+1, $$ref_dict[1]{$input[$ind]}, $depth+1);
	}
	else {
		$$ref_dict[1]{$input[$ind]}[0] = ++$$ref_counter;
		$$ref_dict[1]{$input[$ind]}[1] = {};
		return ($depth+1, $$ref_dict[0], $input[$ind]);
	}
}


# decoding LZW

sub lzw_decode
{
# in:
# encoded input (ref array)
# mode (scalar)
# ref_alphabet (filled)
# out:
# decoded (array)

	my $ref_encoded = shift @_;
	my $mode = shift;
	my $ref_alphabet = shift;

	my @encoded = @$ref_encoded;
	my @decoded = ();
	my @dict = ();
	my $counter = $start;

	$dict[0] = $counter;
	foreach my $a (@$ref_alphabet) {
		$dict[1]{$a}[0] = ++$counter; 
		$dict[1]{$a}[1] = {}; 
	}
	
	my $word = "";
	print_dict_list(\@dict);
	foreach $a (@encoded) {
		my $ret = search_dec_lzw($a, \@dict);
		my @out;
		my $tmp_ret;
		my @out_dict;
		if (not defined $ret) {
			$word .= substr($word,0,1);
			$tmp_ret = add_to_dict(\@dict, \$counter, $word, 0);
			@out_dict = split //, substr($word, 0, $tmp_ret, ''); 
			$ret = search_dec_lzw($a, \@dict);
			if (not defined $ret) {
				print "ERROR\n";
				last;
			}
			$word = $ret;
			@out = split (//, $ret);
			push @decoded, @out;
			push @out_dict, substr($ret, 0, 1, '');
		}

		else {
			@out = split (//, $ret);
			push @decoded, @out;
			$word .= $ret;
			$tmp_ret = add_to_dict(\@dict, \$counter, $word, 0);
			@out_dict = split //, substr($word, 0, $tmp_ret+1);
			substr($word, 0, $tmp_ret, '');
		} 

		printf "%2s .. %-6s\t=>\tout: %-6s\t%s\n", $tmp_ret == 0 ? ("","") : ($counter, "@out_dict"), "@out", "($a)" if no_silent($mode);

		my $key = my_wait($mode);
		if (defined $key) {
			print_dict_tree(\@dict) if ($key eq 't');
			print_dict_list(\@dict) if ($key eq 'l');
			if ($key eq 'b') {
				print_dict_tree(\@dict);
				print_dict_list(\@dict);
			}
		}		
		else {
			print_dict_tree(\@dict) if tree_view($mode);
			print_dict_list(\@dict) if list_view($mode);
		}
		
	}
	return (join '', @decoded);
}


sub search_dec_lzw 
{
	my ($a, $ref_dict, $pref) = @_;
	my @fifo = ();
 	my $prefix = "";
	push @fifo, $ref_dict, $prefix;
 	my $node;
	while ($#fifo != -1) {
		$node = shift @fifo;
		$prefix = shift @fifo;
		if ($$node[0] == $a) {
			return $prefix;
		}
		else {
			while (my ($key, $value) = each( %{$$node[1]} )) {
				push @fifo, $value, $prefix . $key;
			}
		}
	}
	return undef;
}


sub add_to_dict
{
	my ($ref_dict, $ref_counter, $word, $depth) = @_;
	my @word = split //, $word;
	return (0) if $word eq "";
	my $char = shift @word; 
	if (exists $$ref_dict[1]{$char}) {
		return add_to_dict($$ref_dict[1]{$char}, $ref_counter, join('', @word), $depth+1);
	}
	else {
		$$ref_dict[1]{$char}[0] = ++$$ref_counter;
		$$ref_dict[1]{$char}[1] = {};
		return ($depth);
	}
}


1;