package LZ78;

#interface
use Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(&lz78_encode &lz78_decode &print_help_lz78 &lz78_count_bits);

#implementation

use strict;
use POSIX;
use Common;

###############################################################################
# LZ78                                                                       ##
###############################################################################

sub print_help_lz78
{
	print "\nLZ78\n\
Metoda LZ78 je založena na rostoucím slovníku. Text zpracováváme zleva, hledáme \
nejdelší prefix, který je již obsažen ve slovníku, přidáme k němu další písmeno \
ze vstupu a tento řetězec uložíme do slovníku. Na výstup posíláme dvojice (i, z), \
kde 'i' je index do slovníku, kde se nalézá prefix a 'z' je přidané písmeno. \
Je zřejmé, že slovník roste velmi rychle. Lze ho řídit tak, že pokud se přeplní \
místo určené pro slovník, pokračuje se buď s existujícím slovníkem nebo se vymaže \
a začne se budovat znovu.\
V tomto programu je slovník implementován jako strom. Pokud je nastavena volba \
tree_view resp. list_view, po každém kroku bude vypsán jako strom resp. jako \
seznam (v případě, že není nastaven silent_mode). V případě volby waiting_mode \
lze vypisovat slovník explicitně písmeny t (tree_view), l (list_view) nebo \
b (both - vypíše se strom i seznam).\n\n"; 
}

sub lz78_count_bits
{
# in:
# encoded (ref array)
# out:
# count of bits
	my @encoded = @{shift @_};
	my $bits = 0;
	my $max = 3; # starts with 3 bits
	foreach my $code_word (@encoded) {
		my @code_word = split /_/, $code_word;
		my $pom += $code_word[0] == 0 ? 1 : ceil(log($code_word[0])/log(2));
		if ($pom > $max) {
			$max = $pom;
		}	
		$bits += $max;
		$bits += 8;
	}
	return $bits;
}

# encoding LZ78

sub lz78_encode
{
# in:
# input (ref array)
# mode (scalar)
# out:
# encoded (array)

	my $ref_input = shift @_;
	my $mode = shift;
	my @input = split //, $$ref_input;
	my @encoded = ();

	my $input_length = $#input + 1;

# struct of dictionary:
# node consists of two items: number of node and hash (keys: letters, values: nodes)

	my @dict = ();
	my $counter = 0;
	$dict[0] = $counter;
	$dict[1] = {};
	
	for (my $i = 0; $i < $input_length; ) {
		my @ret = search_enc_lz78(\@input, \$counter, $i, \@dict, 0);
		my $out = $ret[1].'_'.$ret[2];
		push @encoded, $out;
		printf "%-6s\t=>\t%7s\n", "@input[$i .. $i+$ret[0]-1]", translate($out) if no_silent($mode); 
		my $key = my_wait($mode);
		if (defined $key) {
			print_dict_tree(\@dict) if ($key eq 't');
			print_dict_list(\@dict) if ($key eq 'l');
			if ($key eq 'b') {
				print_dict_tree(\@dict);
				print_dict_list(\@dict);
			}
		}
		else {
			print_dict_tree(\@dict) if tree_view($mode);
			print_dict_list(\@dict) if list_view($mode);
		}
		
		$i += $ret[0];
	}
	return @encoded;
}

sub search_enc_lz78
{
	my ($ref_input, $ref_counter, $ind, $ref_dict, $depth) = @_;
	my @input = @$ref_input;
	my $input_length = $#input + 1;
	return ($depth, $$ref_dict[0], '') if $ind >= $input_length;
	if (exists $$ref_dict[1]{$input[$ind]}) {
		return search_enc_lz78($ref_input, $ref_counter, $ind+1, $$ref_dict[1]{$input[$ind]}, $depth+1);
	}
	else {
		$$ref_dict[1]{$input[$ind]}[0] = ++$$ref_counter;
		$$ref_dict[1]{$input[$ind]}[1] = {};
		return ($depth+1, $$ref_dict[0], $input[$ind]);
	}
}


# decoding LZ78

sub lz78_decode 
{ 
# in:
# encoded input (ref array)
# mode (scalar)
# out:
# decoded (array)

	my $ref_encoded = shift @_;
	my $mode = shift;

	my @encoded = @$ref_encoded;
	my @decoded = ();
	my @dict = ();
	my $counter = 0;

	$dict[0] = $counter;
	$dict[1] = {};
	foreach $a (@encoded) {
		my @code = split /_/, $a;
		my $ret = search_dec_lz78(\@code, \$counter, \@dict);
		my @out = split (//, $ret);
		push @decoded, @out;
		printf "%-7s  =>  %-6s  =>  %s\n", translate($a), "@out", "@decoded" if no_silent($mode);
		my $key = my_wait($mode);
		if (defined $key) {
			print_dict_tree(\@dict) if ($key eq 't');
			print_dict_list(\@dict) if ($key eq 'l');
			if ($key eq 'b') {
				print_dict_tree(\@dict);
				print_dict_list(\@dict);
			}
		}		
		else {
			print_dict_tree(\@dict) if tree_view($mode);
			print_dict_list(\@dict) if list_view($mode);
		}
	}
	return (join '', @decoded);
}	

sub search_dec_lz78 
{

	my ($ref_code, $ref_counter, $ref_dict) = @_;
	
	my @fifo = ();
 	my $prefix = "";
	push @fifo, $ref_dict, $prefix;
 	my $node;
	while ($#fifo != -1) {
		$node = shift @fifo;
		$prefix = shift @fifo;
		if ($$node[0] == $$ref_code[0]) {
			# prefix found
			$$ref_counter++;
			if (defined $$ref_code[1]) {
				$$node[1]{$$ref_code[1]}[0] = $$ref_counter;
				$$node[1]{$$ref_code[1]}[1] = {};
				return $prefix . $$ref_code[1];
			}
			else {
				return $prefix;
			}
		}
		else {
			while (my ($key, $value) = each( %{$$node[1]} )) {
				push @fifo, $value, $prefix . $key;
			}
		}
	}
}

1;